// Copyright(c) 1996 Leendert Ammeraal. All rights reserved.
// This program text occurs in Chapter 6 of
//
//    Ammeraal, L. (1996) Algorithms and Data Structures in C++,
//       Chichester: John Wiley.

// avl: Demonstration program for AVL trees;
//      insertion and deletion of nodes
#include <iostream.h>
#include <iomanip.h>
#include <ctype.h>

struct node{int num, bal; struct node *pLeft, *pRight;};

class AVLtree {
public:
   AVLtree():root(NULL){}
   void insert(int x){ins(root, x);}
   void DelNode(int x){del(root, x);}
   void print()const{pr(root, 0);}
private:
   node *root;
   void LeftRotate(node* &p);
   void RightRotate(node* &p);
   int ins(node* &p, int x);
   int del(node* &p, int x);
   void pr(const node *p, int nSpace)const;
};

void AVLtree::LeftRotate(node* &p)
{  node *q = p;
   p = p->pRight;
   q->pRight = p->pLeft;
   p->pLeft = q;
   q->bal--;
   if (p->bal > 0) q->bal -= p->bal;
   p->bal--;
   if (q->bal < 0) p->bal += q->bal;
}

void AVLtree::RightRotate(node* &p)
{  node *q = p;
   p = p->pLeft;
   q->pLeft = p->pRight;
   p->pRight = q;
   q->bal++;
   if (p->bal < 0) q->bal -= p->bal;
   p->bal++;
   if (q->bal > 0) p->bal += q->bal;
}

int AVLtree::ins(node* &p, int x)
{  // Return value: increase in height (0 or 1) after
   // inserting x in the (sub)tree with root p
   int deltaH=0;
   if (p == NULL)
   {  p = new node;
      p->num = x; p->bal = 0;
      p->pLeft = p->pRight = NULL;
      deltaH = 1; // Tree height increased by 1
   }  else
   if (x > p->num)
   {  if (ins(p->pRight, x))
      {  p->bal++; // Height of right subtree increased 
         if (p->bal == 1) deltaH = 1; else
         if (p->bal == 2)
         {  if (p->pRight->bal == -1) RightRotate(p->pRight);
            LeftRotate(p);
         }
      }
   }  else
   if (x < p->num)
   {  if (ins(p->pLeft, x))
      {  p->bal--; // Height of left subtree increased 
         if (p->bal == -1) deltaH = 1; else
         if (p->bal == -2)
         {  if (p->pLeft->bal == 1) LeftRotate(p->pLeft);
            RightRotate(p);
         }
      }
   }
   return deltaH;
}

int AVLtree::del(node* &p, int x)
/* Return value: decrease in height (0 or 1) of subtree
   with root p, after deleting the node with key x.
   (If there is no such node, 0 will be returned.)
*/
{  node **qq, *p0;
   int deltaH=0;
   if (p == NULL) return 0;
   if (x < p->num)
   {  if (del(p->pLeft, x))
      {  p->bal++; // Height left subtree decreased
         if (p->bal == 0) deltaH = 1; else
         if (p->bal == 2)
         {  if (p->pRight->bal == -1) RightRotate(p->pRight);
            LeftRotate(p);
            if (p->bal == 0) deltaH = 1;
         }
      }
   }  else
   if (x > p->num)
   {  if (del(p->pRight, x))
      {  p->bal--; // Height right subtree decreased
         if (p->bal == 0) deltaH = 1; else
         if (p->bal == -2)
         {  if (p->pLeft->bal == 1) LeftRotate(p->pLeft);
            RightRotate(p);
            if (p->bal == 0) deltaH = 1;
         }
      }
   }  else  // x == p->num
   {  if (p->pRight == NULL)
      {  p0 = p; p = p->pLeft; delete p0; return 1;
      }  else
      if (p->pLeft == NULL)
      {  p0 = p; p = p->pRight; delete p0; return 1;
      }  else
      {  qq = & p->pLeft;
         while ((*qq)->pRight != NULL) qq = & (*qq)->pRight;
         p->num = (*qq)->num;
         (*qq)->num = x;
         if (del(p->pLeft, x))
         {  p->bal++; // Height left subtree decreased
            if (p->bal == 0) deltaH = 1; else
            if (p->bal == 2)
            {  if (p->pRight->bal == -1) RightRotate(p->pRight);
               LeftRotate(p);
               if (p->bal == 0) deltaH = 1;
            }
         }
      }
   }
   return deltaH;
}

void AVLtree::pr(const node *p, int nSpace)const
{  if (p != NULL)
   {  pr(p->pRight, nSpace+=6);
      cout << setw(nSpace) << p->num << " " << p->bal << endl;
      pr(p->pLeft, nSpace);
   }
}

int main()
{  int x;
   char ch;
   AVLtree t;
   cout <<
   "\nEach AVL tree displayed by this program has\n"
   "its root on the left. Turn it 90 degrees\n"
   "clockwise to obtain its usual representation.\n\n";
   for ( ; ; )
   {  cout <<
      "Enter an integer, followed by I for insertion\n"
      "or by D for deletion, or enter Q to quit: ";
      cin >> x >> ch;
      if (cin.fail()) break;
      ch = toupper(ch);
      if (ch == 'I') t.insert(x); else
      if (ch == 'D') t.DelNode(x);
      t.print();
   }
   return 0;
}		

