// Copyright(c) 1996 Leendert Ammeraal. All rights reserved.
// This program text occurs in Chapter 6 of
//
//    Ammeraal, L. (1996) Algorithms and Data Structures in C++,
//       Chichester: John Wiley.

// infsys: An information system based on a perfectly balanced
//         binary tree.
#include <fstream.h>
#include <iomanip.h>
#include <ctype.h>
#include <string.h>

const int bufLen = 31;
struct node {char *name; long num; node *pLeft, *pRight;};

class tree {
public:
   tree(){root = NULL; ItemCount = 0;}
   tree(const char *FileName);
   ~tree();
   void print()const{pr(root);}
   node* &search(const char *name);
   void save(const char *FileName)const;
   void AddOrChange(const char *name, long nr);
   int DelNode(const char *name);
private:
   node *root;
   node *BuildTree(ifstream &iFile, int n);
   void pr(const node *p)const;
   void DelTree(node *p);
   void WriteTree(ofstream &oFile, const node *p)const;
   int ItemCount;
};

node *tree::BuildTree(ifstream &iFile, int n)
{  if (n == 0) return NULL;
   int nLeft = (n - 1)/2, nRight = n - nLeft - 1;
   char buf[bufLen];
   node *p = new node; ItemCount++;
   p->pLeft = BuildTree(iFile, nLeft);
   iFile >> setw(bufLen) >> buf >> p->num;
   p->name = new char[strlen(buf) + 1];
   strcpy(p->name, buf);
   p->pRight = BuildTree(iFile, nRight);
   return p;
}

tree::tree(const char *FileName)
{  ifstream iFile(FileName, ios::in);
   if (iFile.fail())
   {  cout << "Cannot open input file.\n";
      root = NULL; ItemCount = 0;
   }  else
   {  int n;
      iFile >> n; ItemCount = 0;
      root = BuildTree(iFile, n);
      iFile.close();
   }
}

void tree::DelTree(node *p)
{  if (p)
   {  DelTree(p->pLeft);
      DelTree(p->pRight);
      delete[] p->name;
      delete p;
   }
}

tree::~tree()
{  DelTree(root);
}

void tree::pr(const node *p)const
{  if (p)
   {  pr(p->pLeft);
      cout << setiosflags(ios::left) << setw(bufLen)
           << p->name << resetiosflags(ios::left)
           << setw(8) << p->num << endl;
      pr(p->pRight);
   }
}

node* &tree::search(const char *name)
{  int code;
   node **p = &root; // p is a pointer to a pointer
   for(;;)
   {  if (*p == NULL) return *p;
      code = strcmp(name, (*p)->name);
      if (code < 0) p = &(*p)->pLeft; else
      if (code > 0) p = &(*p)->pRight; else return *p;
   }
}

void tree::WriteTree(ofstream &oFile, const node *p)const
{  if (p)
   {  WriteTree(oFile, p->pLeft);
      oFile << p->name << " " << p->num << endl;
      WriteTree(oFile, p->pRight);
   }
}

void tree::save(const char *FileName)const
{  ofstream oFile(FileName, ios::out);
   if (oFile.fail())
   {  cout << "Cannot open output file.\n"; return;
   }
   oFile << ItemCount << endl;
   WriteTree(oFile, root);
   oFile.close();
}

void tree::AddOrChange(const char *name, long nr)
{  node* &p = search(name);
   if (p == NULL) // Insert new node
   {  p = new node;
      p->name = new char[strlen(name) + 1];
      strcpy(p->name, name);
      p->pLeft = p->pRight = NULL;
      ItemCount++;
   }
   p->num = nr;
}

int tree::DelNode(const char *name)
{  node* &p = search(name), *p0 = p, **qq, *q;
   if (p == NULL) return 0; // Not found
   delete[] p->name;
   if (p->pRight == NULL){p = p->pLeft; delete p0;} else
   if (p->pLeft == NULL){p = p->pRight; delete p0;} else
   {  // Node *p has a left and a right child
      qq = & p->pLeft;
      while ((*qq)->pRight) qq = & (*qq)->pRight;
      // qq contains the address of the pointer field
      // that points to a node without a right child.
      p->name = (*qq)->name; p->num = (*qq)->num;
      // Data copied from **qq to *p
      q = *qq;
      *qq = q->pLeft;
      // *qq replaced with left link field of **qq.
      delete q;
   }
   ItemCount--;
   return 1; // Node deleted
}

int main()
{  char ch, buf[bufLen];
   long nr;
   node *p;
   ifstream iFile;
   ofstream oFile;
   tree *pTree = new tree;
   cout << "Enter a name and a nonnegative integer, or\n"
           "a name followed either by / or by ?, or\n"
           "one of the commands .Load, .Print, .Save, .Quit.\n"
           "There must be a blank space after a name, not\n"
           "within it.\n";
   int busy = 1;
   while (busy)
   {  cout << ">>";
      cin >> setw(bufLen) >> buf;
      if (buf[0] == '.')
      switch (toupper(buf[1]))
      {
      case 'L':
         cout << "Input file: ";
         cin >> setw(bufLen) >> buf;
         delete pTree;
         pTree = new tree(buf);
         break;
      case 'P':
         cout << "Contents:\n";
         pTree->print();
         break;
      case 'Q':
         busy = 0; delete pTree;
         break;
      case 'S':
         cout << "Output file: ";
         cin >> setw(bufLen) >> buf;
         pTree->save(buf);
         break;
      default:
         cout <<
         "Wrong command, use .Load, .Print, .Save, or .Quit\n";
      }  else
      {  cin >> nr;
         if (cin.fail())
         {  cin.clear();
            cin >> ch;
            switch(ch)
            {
            case '?':
               p = pTree->search(buf);
               if (p)
                  cout << setiosflags(ios::left) << setw(bufLen)
                       << p->name << resetiosflags(ios::left)
                       << setw(8) << p->num << endl;
               else cout << "Unknown name.\n";
               break;
            case '/':
               if (!pTree->DelNode(buf))
                  cout << "Unknown name.\n";
               break;
            default:
               cout << "A name must be followed by a number,\n"
                       "a question mark or a slash.\n";
            }
         }  else pTree->AddOrChange(buf, nr);
      }
      do cin.get(ch); while (ch != '\n');
   }
   return 0;
}

