// Copyright(c) 1996 Leendert Ammeraal. All rights reserved.
// This program text occurs in Chapter 8 of
//
//    Ammeraal, L. (1996) Algorithms and Data Structures in C++,
//       Chichester: John Wiley.

// huffman1: File compression (to be linked with huffman.cpp)
#include <limits.h>
#include "huffman.h"

struct TableItem {unsigned code, len;}; // For code table

void traverse(HuffmanTree &Tr, TableItem *t, int p,
   unsigned code, int len)
{  if (Tr.A[p].left == nil)
   {  uchar ch = Tr.A[p].RightOrCh;
      t[ch].code = code;
      t[ch].len = len;
   }  else
   {  code <<= 1; len++;
      traverse(Tr, t, Tr.A[p].left, code, len);
      traverse(Tr, t, Tr.A[p].RightOrCh, code | 1, len);
   }
}

uchar FindLenCode(ulong f)
{  return f > USHRT_MAX ? 3 :
          f > UCHAR_MAX ? 2 : (f != 0);
// Code for number of bytes that each frequency takes:
// 00 -> zero, 01 -> one, 10 -> two, 11 -> four
}

void WriteFreqTable(ofstream &ofile, const ulong *freq)
{  /* Frequencies will now be written to ofile in a compact
      form. We deal with the smallest possible range
      freq[jLower], ..., freq[jUpper] that includes all
      nonzero elements of array freq. We divide this sequence
      into consecutive subsequences, each with elements of a
      fixed length. Each of these subsequences is preceded 
      by a control byte, consisting of a 2-bit length code
      followed by six bits indicating how many of such
      frequencies follow in that sequence.
   */
   int jLower = 0, jUpper = 255, j, j0, LenCode;
   ulong f;
   ushort u;
   while (jLower < 256 && freq[jLower] == 0) jLower++;
   while (freq[jUpper] == 0) jUpper--;
   ofile.put((uchar) jLower);
   ofile.put((uchar) jUpper);
   for (j0 = jLower; j0 <= jUpper; )
   {  LenCode = FindLenCode(freq[j0]);
      j = j0 + 1;
      while (j<=jUpper && j - j0 < 0x3F &&
         FindLenCode(freq[j]) == LenCode) j++;
      ofile.put(uchar((LenCode << 6) | (j - j0)));
      while (j0 < j)
      {  f = freq[j0++];
         switch(LenCode){
         case 1: ofile.put(uchar(f)); break;
         case 2: u = ushort(f);
                 ofile.write((char*)&u, sizeof(ushort));
                 break;
         case 3: ofile.write((char*)&f, sizeof(ulong));
                 break;
         }
      }
   }
}

int main()
{  ifstream ifile;
   ofstream ofile;
   cout << "Huffman file compression.\n";
   GetStreams(ifile, ofile);
   ulong freq[256];
   uchar ch;
   int j, k, b;
   for (j=0; j<256; j++) freq[j] = 0;
   while (ifile.get(ch), !ifile.fail())
      freq[ch]++;
   ifile.clear(); ifile.seekg(0, ios::beg);

   HuffmanTree Tr(freq);

   TableItem t[256]; // t is the code table
   for (int i=0; i<256; i++) t[i].code = t[i].len = 0;
   traverse(Tr, t, Tr.root, 0, 0);
   WriteFreqTable(ofile, freq);
   uchar buf = 0;
   k = 0;
   while (ifile.get(ch), !ifile.fail())
   {  for (int i=t[ch].len-1; i>=0; i--)
      {  b = (t[ch].code >> i) & 1;
         if (k < 8){buf = (buf << 1) | b; k++;} else
         {  ofile.put(buf); buf = b; k = 1;
         }
      }
   }
   ofile.put(char(buf << (8 - k)));
   ofile.put(char(k)); // k relevant bits in last byte
   return 0;
}

