// Copyright(c) 1996 Leendert Ammeraal. All rights reserved.
// This program text occurs in Chapter 8 of
//
//    Ammeraal, L. (1996) Algorithms and Data Structures in C++,
//       Chichester: John Wiley.

// trie: Demonstration program for tries
#include <stdlib.h>
#include <fstream.h>
#include <iomanip.h>
#include <string.h>
#include <ctype.h>

const int MaxWordLength = 80;
const int nBranches = 'Z' - 'A' + 2; // = 27 with ASCII
struct dnode{int n;  char *s;};
typedef dnode *ptr;
struct bnode{int n; ptr p[nBranches];};
typedef bnode *bPtr;

class trie {
public:
   trie(const char *fname);
   void insert(const char *s);
   void print()
   {  if (root){cout << "Contents of trie:\n"; pr(root);} 
      else cout << "Empty trie.\n";
   }
   ptr search(const char *s);
private:
   void pr(const ptr r);
   ptr NewNode(const char *s = NULL);
   int position(char ch){return ch ? ch - 'A' + 1 : 0;}
   ptr root;
};

int GetWord(istream &f, char *s)
{  char ch;
   int i;
   do
   {  f >> ch; ch = toupper(ch);
      if (f.fail()) return 0;
   }  while (ch < 'A' || ch > 'Z');
   for (i=0; i<MaxWordLength-1; )
   {  s[i++] = ch;
      f.get(ch); ch = toupper(ch);
      if (f.fail() || ch < 'A' || ch > 'Z') break;
   }
   s[i] = '\0';
   return i;
}

trie::trie(const char *fname)
{  char s[MaxWordLength + 1];
   ifstream file(fname, ios::in | ios::nocreate);
   if (file.fail())
   {  cout << "Cannot open input file.\n";
      exit(1);
   }
   root = NULL;
   while (GetWord(file, s) > 0) insert(s);
}

void trie::insert(const char *s)
{  int m = 0;
   ptr *pp = &root, q;
     // pp contains the address of the current pointer
     // q will be a copy of this current pointer
   for (;;)
   {  q = *pp;
      if (q == NULL){*pp = NewNode(s); return;}
      if (q->n > 0)  // Data node:
      {  if (strcmp(s, q->s) == 0){q->n++; return;}
         // Prefix s[0], ..., s[m-1] of new word identical
         // with that of existing word q->s.
         // Insert a new branch node between pointer
         // *pp and data node *q, and continue at next level:
         ptr q0 = q;
         *pp = q = NewNode();
         int j = position(q0->s[m]);
         bPtr(q)->p[j] = q0;
      }
      int i = position(s[m++]);
      pp = & bPtr(q)->p[i];  // Go to next level to insert s
   }
}

void trie::pr(const ptr r)
{  if (r->n)    // n > 0: r points to data node
      cout << setw(5) << r->n << " " << r->s << endl;
   else         // n = 0: r points to a branch node
   for (int i=0; i<nBranches; i++)
   {  ptr q = bPtr(r)->p[i];
      if (q) pr(q);
   }
}

ptr trie::NewNode(const char *s)
{  ptr q;
   if (s) // New data node:
   {  q = new dnode;
      q->n = 1;
      q->s = new char[strlen(s) + 1];
      strcpy(q->s, s);
   }  else // New branch node:
   {  bPtr qb = new bnode;
      qb->n = 0;
      for (int i=0; i<nBranches; i++) qb->p[i] = NULL;
      q = ptr(qb);
   }
   return q;
}

ptr trie::search(const char *s)
{  ptr r = root;
   char ch;
   int m=0, i;
   while (r && r->n == 0) // While r points to branch node
   {  ch = s[m++];
      i = position(ch);
      if (i < 0 || i >= nBranches) return NULL;
      r = bPtr(r)->p[i];
   }
   return r && strcmp(s, r->s) == 0 ? r : NULL;
}

int main()
{  char fname[50], s[MaxWordLength + 1];
   ptr q;
   cout <<
      "This program can read any text file. Then for any\n"
      "word (consisting of at most " << MaxWordLength
      << " letters), you can\n"
      "ask how often that word occurs in this file.\n"
      "Input file: ";
   cin >> setw(50) >> fname;
   trie t(fname); // Create entire trie
   t.print();
   for (;;)
   {  cout << "Enter a word, or type Ctrl+Z (or Ctrl+D) "
         "to stop: ";
      if (GetWord(cin, s) == 0) break;
      q = t.search(s);
      if (q) cout << "Word found, count = " << q->n << endl;
      else cout << "Word not found.\n";
   }
   return 0;
}

