// Copyright(c) 1996 Leendert Ammeraal. All rights reserved.
// This program text occurs in Chapter 9 of
//
//    Ammeraal, L. (1996) Algorithms and Data Structures in C++,
//       Chichester: John Wiley.

// shpath: Finding the shortest path: Dijkstra's algorithm
//   (to be linked with strings.cpp, see Section 9.5).
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>
#include "strings.h"

struct edge{int num, len; edge *next;};

class graph {
public:
   int n, N, start, finish, nInS, undirected;
   void build(strings &T, ifstream &file);
   void output(const strings &T)const;
   int length(int i, int j)const;
   void prepare();
   void ExpandSetS();
private:
   struct NodeType{int count, D, connect, inS; edge *link;} *list;
   enum {nil=-1, inf=INT_MAX/2};
   int ReadName(strings &T, istream &file, char &PrevChar);
};

void error(const char *s)
{  cout << s << endl; exit(1);
}

int graph::ReadName(strings &T, istream &file, char &PrevChar)
{  char s[100];
   s[0] = '\n';
   do
   {  PrevChar = s[0];
      file.get(s[0]);
   }  while (!file.fail() && !isalpha(s[0]));
   if (file.fail()) return -1;
   file >> setw(99) >> &s[1];
   return T.add(s);
}

void graph::build(strings &T, ifstream &file)
{  int i, j, len;
   edge *p;
   char PrevChar;
   const int ChunkSize = 64;
   N = n = 0;
   list = NULL;
   for (;;)
   {  j = ReadName(T, file, PrevChar);
      if (j < 0) break;
      if (PrevChar == '\n')
      {  i = j;
         j = ReadName(T, file, PrevChar);
      }
      file >> len;
      if (file.eof()) break;
      if (file.fail()) error("Incorrect input file.");
      int max = i > j ? i : j;
      if (max >= N)
      {  NodeType *listOld = list;
         list = new NodeType[N += ChunkSize];
         for (int v=0; v<N; v++)
            if (v < n) list[v] = listOld[v]; else
            {  list[v].count = 0;
               list[v].link = NULL;
            }
         delete[]listOld;
      }
      if (max >= n) n = max + 1;
      p = new edge;
      p->num = j; p->len = len;
      p->next = list[i].link;
      list[i].link = p;
      list[j].count++;
             // Undirected graph: add <j, i> besides <i, j> :
      if (undirected)
      {  p = new edge;
         p->num = i; p->len = len;
         p->next = list[j].link;
         list[j].link = p;
         list[i].count++;
      }
   }
   file.close();
}

int graph::length(int i, int j)const
{  edge *p = list[i].link;
   while (p)
   {  if (p->num == j) return p->len;
      p = p->next;
   }
   return inf;
}

void graph::prepare()
{  for (int v=0; v<n; v++)
   {  list[v].D = (v == start ? 0 : length(start, v));
      list[v].inS = 0;
      list[v].connect = start;
   }
   list[start].inS = 1;
   nInS = 1;
}

void graph::ExpandSetS()
{  int v, w = inf, min = INT_MAX;
   // Choose a vertex w in V such that list[w].D
   // is a minimum:
   for (v=0; v<n; v++)
   if (list[v].inS == 0)
   {  if (list[v].D < min)
      {  min = list[v].D;
         w = v;
      }
   }  // w != inf
   // Move w from V to S:
   list[w].inS = 1;  nInS++;
   // Update distances list[v].D and predecessors
   // list[v].connect for all vertices v in V:
   for (v=0; v<n; v++)
   if (list[v].inS == 0)
   {  int sum = list[w].D + length(w, v);
      if (sum < list[v].D)
      {  list[v].D = sum;
         list[v].connect = w;
      }
   }
}

void graph::output(const strings &T)const
{  if (list[finish].D == inf)
   {  cout << "No path from start to finish.\n";
      return;
   }
   // Replace predecessors with successors:
   int j = finish, i, jnext = 0;
   for (;;)
   {  i = list[j].connect;
      list[j].connect = jnext;
      if (j == start) break;
      jnext = j;
      j = i;
   }
   cout << "Shortest path:\n\n";
   i = start;
   cout << "         Vertex              Distance\n";
   for (;;)
   {  cout << (i == finish ? "finish = " :
               i == start  ? "start  = " :
                             "         ")
         << setw(20) << setiosflags(ios::left) << T[i] << "  "
         << setw(6) << resetiosflags(ios::left) << list[i].D 
         << endl;
      if (i == finish) break;
      i = list[i].connect;
   }
}

int main()
{  graph g;
   strings T;
   char fname[50], ch;
   cout << "Input file: ";
   cin >> setw(50) >> fname;
   ifstream file(fname, ios::in | ios::nocreate);
   if (file.fail()) error("Cannot open input file.");
   cout << "Is the graph Directed or Undirected? (D/U) ";
   cin >> ch;
   g.undirected = (ch == 'u' || ch == 'U');
   g.build(T, file);     // Build adjacency lists
   cout << "Vertices:";
   int n = T.size();
   for (int i=0; i<n; i++)
   {  if (i % 5 == 0) cout << endl;
      cout << T[i] << " ";
   }
   do
   {  cout << "\nEnter start and finish:\n";
      g.start = T.read(cin);
      g.finish = T.read(cin);
      if (g.start >= g.n || g.finish >= g.n)
         cout << "Incorrect start or finish.\n"; else
      {  g.prepare();
         while (g.nInS < g.n)
            g.ExpandSetS();
         g.output(T);
      }
      cout << "Do you want another path? (Y/N) ";
      cin >> ch;
   }  while (ch == 'y' || ch == 'Y');
   return 0;
}

